import abc
import logging
from typing import Generic
from typing import Optional
from typing import Type
from typing import TypeVar
from typing import Union

import aiofiles
import aiohttp.web_response
import pydantic
import ujson
from aiohttp.typedefs import PathLike

from .types import EmptyResponse
from .utils import DEFAULT_DOWNLOAD_CHUNK_SIZE

ResponseContentType = TypeVar('ResponseContentType')


class ResponseClass(abc.ABC, Generic[ResponseContentType]):
    charset: str = "utf-8"
    aiohttp_response: aiohttp.ClientResponse

    def __init__(self, aiohttp_response: aiohttp.ClientResponse):
        self.aiohttp_response = aiohttp_response
        self.logger = logging.getLogger(self.__class__.__name__)

    async def parse(self, *args, **kwargs) -> Optional[ResponseContentType]:
        return self.aiohttp_response.content


class RawResponseClass(ResponseClass[aiohttp.ClientResponse]):
    async def parse(self, *args, **kwargs) -> aiohttp.ClientResponse:
        return self.aiohttp_response


class NoneResponseClass(ResponseClass[None]):
    async def parse(self) -> None:
        return None


class PlainTextResponseClass(ResponseClass[str]):
    async def parse(self, *args, **kwargs) -> str:
        return await self.aiohttp_response.text(self.charset)


_JsonBaseFields = Union[str, int, float, bool, None]
Json = Union[_JsonBaseFields, dict[str, _JsonBaseFields], list[_JsonBaseFields]]


class JSONResponseClass(ResponseClass[Json]):
    async def parse(self, *args, **kwargs) -> Optional[Json]:
        return await self.aiohttp_response.json(
            encoding=self.charset,
            loads=ujson.loads,
            content_type=None
        )


PydanticModel = TypeVar('PydanticModel')


class PydanticModelResponseClass(ResponseClass[PydanticModel]):
    async def parse(self, *args, response_model: Type[PydanticModel], **kwargs) -> PydanticModel:
        if response_model is None:
            return EmptyResponse()
            # raise ValueError('response_model could not be None. If you need bare dict use JSONResponseClass instead')

        response_json = await self.aiohttp_response.json(
            encoding=self.charset,
            loads=ujson.loads,
            content_type=None
        )

        return pydantic.parse_obj_as(response_model, response_json)


class StreamResponseClass(ResponseClass[PathLike]):
    async def parse(
            self,
            *args,
            filepath: PathLike,
            chunk_size: int = DEFAULT_DOWNLOAD_CHUNK_SIZE,
            **kwargs
    ) -> PathLike:
        async with aiofiles.open(filepath, 'wb') as fd:
            async for chunk in self.aiohttp_response.content.iter_chunked(chunk_size):
                await fd.write(chunk)

        return filepath
